from __future__ import annotations

import datetime
import logging
import math
from functools import update_wrapper
from typing import TYPE_CHECKING, Iterable

import click

from .state import State
from .types import LayerType, multiple_to_layer_ids, single_to_layer_id

if TYPE_CHECKING:  # pragma: no cover
    from .cli import ProcessorType

__all__ = [
    "layer_processor",
    "global_processor",
    "generator",
    "block_processor",
    "pass_state",
]


def _format_timedelta(dt: datetime.timedelta) -> str:
    s = dt.total_seconds()

    out = ""
    if s >= 60.0:
        minutes = math.floor(s / 60)
        s -= 60 * minutes

        out += str(minutes) + "min "

    out += f"{s:.3f}s"
    return out


def layer_processor(f):
    """Decorator to define a :ref:`layer processor <fundamentals_layer_processors>`
    command.

    Layer processors implements "intra-layer" processing, i.e. they are independently called
    for every layer in the pipeline. A ``--layer`` option is automatically appended to the
    option to let the user control on which layer(s) the processor should be applied (by
    default, ``all`` is used).

    Layer processors receive a :py:class:`LineCollection` as input and must return one.

    Example:

    .. code-block:: python3

        @click.command()
        @layer_processor
        def my_processor(lines: vpype.LineCollection) -> vpype.LineCollection:
            '''Example layer processor'''

            new_lines = vpype.LineCollection()

            for line in lines:
                # [do something with line]
                new_lines.append(line)

            return lines

        my_processor.help_group = "My Plugins"
    """

    @click.option(
        "-l",
        "--layer",
        type=LayerType(accept_multiple=True),
        default="all",
        help="Target layer(s) or 'all'.",
    )
    def new_func(*args, **kwargs):
        layers = kwargs.pop("layer", [])

        # noinspection PyShadowingNames
        def layer_processor(state: State) -> None:
            layers_eval = state.preprocess_argument(layers)

            for lid in multiple_to_layer_ids(layers_eval, state.document):
                start = datetime.datetime.now()
                with state.current():
                    state.current_layer_id = lid
                    new_args, new_kwargs = state.preprocess_arguments(args, kwargs)
                    logging.info(
                        f"executing layer processor `{f.__name__}` on layer {lid} "
                        f"(kwargs: {new_kwargs})"
                    )
                    state.document[lid] = f(state.document[lid], *new_args, **new_kwargs)
                    state.current_layer_id = None
                stop = datetime.datetime.now()

                logging.info(
                    f"layer processor `{f.__name__}` execution complete "
                    f"({_format_timedelta(stop - start)})"
                )

        return layer_processor

    return update_wrapper(new_func, f)


def global_processor(f):
    """Decorator to define a :ref:`global processor <fundamentals_global_processors>`
    command.

    This type of command implement a global, multi-layer processing and should be used for
    processors which cannot be applied layer-by-layer independently (in which case, using
    a :func:`layer_processor` is advised).

    No option is automatically added to global processors. In cases where the user should be
    able to control on which layer(s) the processing must be applied, it is advised to
    add a ``--layer`` option (with type :py:class:`LayerType`) and use the
    :func:`multiple_to_layer_ids` companion function (see example below)

    A global processor receives a :py:class:`Document` as input and must return one.

    Example::

        @click.command()
        @click.option(
            "-l",
            "--layer",
            type=vpype.LayerType(accept_multiple=True),
            default="all",
            help="Target layer(s).",
        )
        @global_processor
        def my_global_processor(
            document: vpype.Document, layer: Union[int, List[int]]
        ) -> vpype.Document:
            '''Example global processor'''

            layer_ids = multiple_to_layer_ids(layer, document)
            for lines in document.layers_from_ids(layer_ids):
                # [apply some modification to lines]

            return document


        my_global_processor.help_group = "My Plugins"
    """

    def new_func(*args, **kwargs):
        # noinspection PyShadowingNames
        def global_processor(state: State) -> None:
            start = datetime.datetime.now()
            with state.current():
                new_args, new_kwargs = state.preprocess_arguments(args, kwargs)
                logging.info(
                    f"executing global processor `{f.__name__}` (kwargs: {new_kwargs})"
                )
                state.document = f(state.document, *new_args, **new_kwargs)
            stop = datetime.datetime.now()

            logging.info(
                f"global processor `{f.__name__}` execution complete "
                f"({_format_timedelta(stop - start)})"
            )

        return global_processor

    return update_wrapper(new_func, f)


def generator(f):
    """Decorator to define a :ref:`generator <fundamentals_generators>` command.

    Generator do not have input, have automatically a "-l, --layer" option added to them, and
    must return a LineCollection structure, which will be added to a new layer or an existing
    one depending on the option.
    """

    @click.option(
        "-l",
        "--layer",
        type=LayerType(accept_new=True),
        default=None,
        help="Target layer or 'new'.",
    )
    def new_func(*args, **kwargs):
        layer = kwargs.pop("layer", None)

        # noinspection PyShadowingNames
        def generator(state: State) -> None:
            with state.current():
                layer_eval = state.preprocess_argument(layer)
                target_layer = single_to_layer_id(layer_eval, state.document)

                start = datetime.datetime.now()
                state.current_layer_id = target_layer
                new_args, new_kwargs = state.preprocess_arguments(args, kwargs)
                logging.info(
                    f"executing generator `{f.__name__}` to layer {target_layer} "
                    f"(kwargs: {new_kwargs})"
                )
                state.document.add(f(*new_args, **new_kwargs), target_layer)
                state.current_layer_id = None
                stop = datetime.datetime.now()

            state.target_layer_id = target_layer

            logging.info(
                f"generator `{f.__name__}` execution complete "
                f"({_format_timedelta(stop - start)})"
            )

        return generator

    return update_wrapper(new_func, f)


def block_processor(f):
    """Decorator to define a `block <fundamentals_blocks>`_ processor command.

    Block processor function take a :class:`State` instance and a processor list as their
    first two argument. The processor list consists of the commands enclosed within the block.
    The block processor implementation should call the :func:`execute_processors` function
    to run the processor list, possibly temporarily changing the content of the :class:`State`
    instance content.

    Example::

        @click.command()
        @vpype_cli.block_processor
        def my_block_processor(state: vpype_cli.State, processors) -> vpype_cli.State:
            '''Example block processor'''

            # block implementation
            # should include call(s) to vpype_cli.execute_processors(processors, state)

            return state

        my_block_processor.help_group = "My Plugins"
    """

    def new_func(*args, **kwargs):
        # noinspection PyShadowingNames
        def block_processor(state: State, processors: Iterable[ProcessorType]) -> None:
            start = datetime.datetime.now()
            new_args, new_kwargs = state.preprocess_arguments(args, kwargs)
            logging.info(f"executing block processor `{f.__name__}` (kwargs: {new_kwargs})")
            f(state, processors, *new_args, **new_kwargs)
            stop = datetime.datetime.now()

            logging.info(
                f"block processor `{f.__name__}` execution complete "
                f"({_format_timedelta(stop - start)})"
            )

        # mark processor as being a block processor, needed by execute_processors()
        block_processor.__vpype_block_processor__ = True
        return block_processor

    return update_wrapper(new_func, f)


def pass_state(f):
    """Marks a command as wanting to receive the current state.

    Note: :func:`pass_state` must always be placed *after* the command type decorator (e.g.
    :func:`generator` and friends).

    Example::

        @click.command()
        @generator
        @pass_state
        def my_generator(state: State) -> vpype.Document:
            lc = vpype.LineCollection()
            current_layer_id = state.current_layer_id
            # ...

            return lc
    """

    def new_func(*args, **kwargs):
        return f(State.get_current(), *args, **kwargs)

    return update_wrapper(new_func, f)



from __future__ import annotations

import math

import numpy as np

__all__ = ["line_length", "is_closed", "interpolate", "crop_half_plane", "crop", "reloop"]


def line_length(line: np.ndarray) -> float:
    """Compute the length of a line."""
    return float(np.sum(np.abs(np.diff(line))))


def is_closed(line: np.ndarray, tolerance: float) -> bool:
    """Check if a line is closed.

    Args:
        line: the line to test
        tolerance: max distance between starting and ending point to consider a path is_closed

    Returns:
        True if line is is_closed
    """
    return len(line) > 1 and np.abs(line[-1] - line[0]) <= tolerance


def interpolate(line: np.ndarray, step: float) -> np.ndarray:
    """Compute a linearly interpolated version of `line` with segments of `step` length or
    less.

    Args:
        line: 1D array of complex
        step: maximum length of interpolated segment

    Returns:
        interpolated 1D array of complex
    """

    curv_absc = np.cumsum(np.hstack([0, np.abs(np.diff(line))]))
    return np.interp(
        np.linspace(0, curv_absc[-1], 1 + math.ceil(curv_absc[-1] / step)), curv_absc, line
    )


def _interpolate_crop(start: complex, stop: complex, loc: float, axis: int) -> complex:
    """Interpolate between two points at a given coordinates."""

    start_dim, stop_dim = (start.real, stop.real) if axis == 0 else (start.imag, stop.imag)

    diff = stop_dim - start_dim
    if diff == 0:
        raise ValueError("cannot interpolate line parallel to axis")

    r = (loc - start_dim) / diff

    if r < 0.5:
        return start + (stop - start) * r
    else:
        return stop - (stop - start) * (1.0 - r)


def crop_half_plane(
    line: np.ndarray, loc: float, axis: int, keep_smaller: bool
) -> list[np.ndarray]:
    """Crop a path at a axis-aligned location.

    The path is cut at location x=loc for axis=0, and y=loc for axis=1. The argument
    `keep_smaller` controls which part of the path is discarded.

    Args:
        line: path to crop
        loc: coordinate at which the cut is made
        axis: 0 for a cut along x axis, 1 for y axis
        keep_smaller: if True, parts of `line` with coordinates smaller or equal than `loc`
            are kept, and the reverse otherwise

    Returns:
        list of paths
    """

    if axis not in [0, 1]:
        raise ValueError("axis must be either 0 or 1")

    line_test_dim = line.real if axis == 0 else line.imag
    if keep_smaller:
        outside = line_test_dim > loc
    else:
        outside = line_test_dim < loc

    if np.all(outside):
        return []
    elif np.all(~outside):
        return [line]

    diff = np.diff(outside.astype(int))
    (start_idx,) = np.nonzero(diff == -1)
    (stop_idx,) = np.nonzero(diff == 1)

    # The following properties are expected after normalization:
    #   - len(start_idx) == len(stop_idx)
    #   - (start_idx, stop_idx) are streak of points that must be kept
    #   - start_idx == -1 means beginning of line in not to be cropped
    #   - stop_idx == inf means end of line is not to be cropped
    inf = len(line)
    if len(start_idx) == 0:
        start_idx = np.array([-1])
    if len(stop_idx) == 0:
        stop_idx = np.array([inf])
    if start_idx[0] > stop_idx[0]:
        start_idx = np.hstack([-1, start_idx])
    if stop_idx[-1] < start_idx[-1]:
        stop_idx = np.hstack([stop_idx, inf])

    line_arr = []
    for start, stop in zip(start_idx, stop_idx):
        if start == -1:
            sub_line = np.hstack(
                [line[: stop + 1], _interpolate_crop(line[stop], line[stop + 1], loc, axis)]
            )
        elif stop == inf:
            sub_line = np.hstack(
                [
                    _interpolate_crop(line[start], line[start + 1], loc, axis),
                    line[start + 1 :],
                ]
            )
        else:
            sub_line = np.hstack(
                [
                    _interpolate_crop(line[start], line[start + 1], loc, axis),
                    line[start + 1 : stop + 1],
                    _interpolate_crop(line[stop], line[stop + 1], loc, axis),
                ]
            )

        # check cases where coordinate lie on threshold
        sub_start = 1 if sub_line[0] == sub_line[1] else 0
        sub_stop = (len(sub_line) - 1) if sub_line[-1] == sub_line[-2] else len(sub_line)
        if sub_stop >= sub_start + 2:
            line_arr.append(sub_line[sub_start:sub_stop])

    return line_arr


def _crop_half_plane_mult(lines: list[np.ndarray], loc: float, axis: int, keep_smaller: bool):
    new_lines = []
    for line in lines:
        new_lines.extend(crop_half_plane(line, loc=loc, axis=axis, keep_smaller=keep_smaller))
    return new_lines


def crop(line: np.ndarray, x1: float, y1: float, x2: float, y2: float) -> list[np.ndarray]:
    """Crop a polyline to a rectangular area.

    Args:
        line: line to crop
        x1: left coordinate of the crop area
        y1: bottom coordinate of the crop area
        x2: right coordinate of the crop area
        y2: top coordinate of the crop area

    Returns:
        list of lines resulting of the crop (emtpy if x1 > x2 or y1 > y2)
    """
    line_list = _crop_half_plane_mult([line], x1, axis=0, keep_smaller=False)
    line_list = _crop_half_plane_mult(line_list, x2, axis=0, keep_smaller=True)
    line_list = _crop_half_plane_mult(line_list, y1, axis=1, keep_smaller=False)
    return _crop_half_plane_mult(line_list, y2, axis=1, keep_smaller=True)


def reloop(line: np.ndarray, loc: int | None = None) -> np.ndarray:
    """Change the seam of a closed path. Closed-ness is not checked. Beginning and end points
    are averaged to compute a new point. A new seam location can be provided or will be chosen
    randomly.

    Args:
        line: path to reloop
        loc: new seam location

    Returns:
        re-seamed path
    """

    if loc is None:
        loc = np.random.randint(len(line) - 1)
    end_point = 0.5 * (line[0] + line[-1])
    line[0] = end_point
    line[-1] = end_point
    return np.hstack([line[loc:], line[1 : min(loc + 1, len(line))]])



from __future__ import annotations

import math
import re
from typing import Callable

import numpy as np

# REMINDER: anything added here must be added to docs/api.rst
__all__ = [
    "UNITS",
    "UNIT_SYSTEMS",
    "ANGLE_UNITS",
    "PAGE_SIZES",
    "convert_length",
    "format_length",
    "convert_angle",
    "convert_page_size",
    "union",
]


def _mm_to_px(x: float, y: float) -> tuple[float, float]:
    return x * 96.0 / 25.4, y * 96.0 / 25.4


# NOTE: order matters due to the implementation of _convert_unit()

UNITS = {
    "px": 1.0,
    "in": 96.0,
    "inch": 96.0,
    "ft": 12.0 * 96.0,
    "yd": 36.0 * 96.0,
    "mi": 1760.0 * 36.0 * 96.0,
    "mm": 96.0 / 25.4,
    "cm": 96.0 / 2.54,
    "m": 100.0 * 96.0 / 2.54,
    "km": 100_000.0 * 96.0 / 2.54,
    "pc": 16.0,
    "pt": 96.0 / 72.0,
}

UNIT_SYSTEMS = {
    "metric": ["mm", "cm", "m", "km"],
    "imperial": ["in", "ft", "yd", "mi"],
}

ANGLE_UNITS = {
    "deg": 1.0,
    "grad": 9.0 / 10.0,
    "rad": 180.0 / math.pi,
    "turn": 360.0,
}

# page sizes in pixel
PAGE_SIZES = {
    "tight": _mm_to_px(0, 0),
    "a6": _mm_to_px(105.0, 148.0),
    "a5": _mm_to_px(148.0, 210.0),
    "a4": _mm_to_px(210.0, 297.0),
    "a3": _mm_to_px(297.0, 420.0),
    "a2": _mm_to_px(420.0, 594.0),
    "a1": _mm_to_px(594.0, 841.0),
    "a0": _mm_to_px(841.0, 1189.0),
    "letter": _mm_to_px(215.9, 279.4),
    "legal": _mm_to_px(215.9, 355.6),
    "executive": _mm_to_px(185.15, 266.7),
    "tabloid": _mm_to_px(279.4, 431.8),
}

_FLOAT_WITH_UNIT_RE = re.compile(r"^([+\-0-9.e]*)([a-z]*)$", flags=re.IGNORECASE)


def _convert_unit(value: str | float | int, units: dict[str, float]) -> float:
    """Converts a string with unit to a value"""
    if isinstance(value, (float, int)):
        return value

    mo = _FLOAT_WITH_UNIT_RE.match(value.strip().lower())
    try:
        if mo is None or value == "":
            raise ValueError
        number = mo.groups()[0]
        unit = mo.groups()[1]
        return (float(number) if number else 1.0) * (units[unit] if unit else 1.0)
    except (ValueError, KeyError):
        raise ValueError(f"cannot convert value '{value}'")


def format_length(value: float, unit: str, full_precision: bool = True) -> str:
    """Convert a length in pixel to a string representation with the provided unit or unit
    system.

    The representation is for display only and

    Args:
        value: the value to convert (in pixels)
        unit: unit (from :data:`vpype.UNITS`) or unit system (from :data:`vp.UNIT_SYSTEMS`)
        full_precision: if False, the return string will use a limited precision best suited
            for informative display

    Returns:
        a string representation of the length
    """
    if unit in UNITS:
        units = [unit]
    elif unit in UNIT_SYSTEMS:
        units = UNIT_SYSTEMS[unit]
    else:
        raise ValueError(f"unknown unit '{unit}'")

    scaled_values = zip((value / UNITS[unit] for unit in units), units)
    picked_value, picked_unit = sorted(
        scaled_values, key=lambda x: abs(math.log10(abs(x[0])) - 0.5) if abs(x[0]) > 0 else 0
    )[0]

    if full_precision:
        return str(picked_value) + picked_unit
    elif picked_value < 0.1:
        return f"{picked_value:.4g}" + picked_unit
    else:
        return f"{picked_value:.3f}" + picked_unit


def convert_length(value: str | float) -> float:
    """Convert a length optionally expressed as a string with unit to px value.

    Args:
        value: value to convert

    Returns:
        converted value

    Raises:
        :class:`ValueError`
    """
    return _convert_unit(value, UNITS)


def convert_angle(value: str | float) -> float:
    """Convert an angle optionally expressed as a string with unit to degrees.

    Args:
        value: angle to convert

    Returns:
        converted angle in degree

    Raises:
        :class:`ValueError`
    """
    return _convert_unit(value, ANGLE_UNITS)


def convert_page_size(value: str) -> tuple[float, float]:
    """Converts a string with page size to dimension in pixels.

    The input can be either a known page size (see ``vpype write --help`` for a list) or
    a page size descriptor in the form of "WxH" where both W and H can have units.

    Examples:

        Using a know page size::

            >>> import vpype
            >>> vpype.convert_page_size("a3")
            (1122.5196850393702, 1587.4015748031497)

        Using page size descriptor (no units, pixels are assumed)::

            >>> vpype.convert_page_size("100x200")
            (100.0, 200.0)

        Using page size descriptor (explicit units)::

            >>> vpype.convert_page_size("1inx2in")
            (96.0, 192.0)

    Args:
        value: page size descriptor

    Returns:
        the page size in CSS pixels
    """
    if value in PAGE_SIZES:
        return PAGE_SIZES[value]

    match = re.match(
        r"^(\d+\.?\d*)({0})?x(\d+\.?\d*)({0})?$".format("|".join(UNITS.keys())), value
    )

    if not match:
        raise ValueError(f"page size '{value}' unknown")

    x, x_unit, y, y_unit = match.groups()

    if not x_unit:
        x_unit = y_unit if y_unit else "px"
    if not y_unit:
        y_unit = x_unit

    return float(x) * convert_length(x_unit), float(y) * convert_length(y_unit)


def union(line: np.ndarray, keys: list[Callable[[np.ndarray], bool]]) -> bool:
    """Returns True if every callables in ``keys`` return True (similar to ``all()``. This
    function is typically used with :meth:`LineCollection.filter`.

    Args:
        line: line to test
        keys: list of callables

    Returns:
        True if every callables return True
    """
    for key in keys:
        if not key(line):
            return False
    return True



from __future__ import annotations

import logging
import pathlib
import sys

import click

import vpype as vp

from .cli import cli
from .decorators import global_processor
from .types import LayerType, LengthType, PageSizeType, PathType, single_to_layer_id

__all__ = ("read",)


@cli.command(group="Input")
@click.argument("file", type=PathType(dir_okay=False, allow_dash=True))
@click.option("-m", "--single-layer", is_flag=True, help="Single layer mode.")
@click.option(
    "-l",
    "--layer",
    type=LayerType(accept_new=True),
    help="Target layer or 'new' (single layer mode only).",
)
@click.option(
    "-a",
    "--attr",
    type=str,
    multiple=True,
    help="Attribute by which geometries should be grouped",
)
@click.option(
    "-q",
    "--quantization",
    type=LengthType(),
    default="0.1mm",
    help="Maximum length of segments approximating curved elements (default: 0.1mm).",
)
@click.option("--no-fail", is_flag=True, help="Do not fail is the target file doesn't exist.")
@click.option(
    "-s",
    "--simplify",
    is_flag=True,
    default=False,
    help="Apply simplification algorithm to curved elements.",
)
@click.option(
    "-p",
    "--parallel",
    is_flag=True,
    default=False,
    help="Enable multiprocessing for SVG conversion.",
)
@click.option(
    "-c",
    "--no-crop",
    is_flag=True,
    default=False,
    help="Do not crop the geometries to the SVG boundaries.",
)
@click.option(
    "-ds",
    "--display-size",
    type=PageSizeType(),
    help=(
        "Display size to use for SVG with width/height expressed as percentage or missing "
        "altogether (see `write` command for possible format)."
    ),
)
@click.option(
    "-dl",
    "--display-landscape",
    is_flag=True,
    default=False,
    help="Use landscape orientation for display size.",
)
@global_processor
def read(
    document: vp.Document,
    file,
    single_layer: bool,
    layer: int | None,
    attr: list[str],
    quantization: float,
    no_fail: bool,
    simplify: bool,
    parallel: bool,
    no_crop: bool,
    display_size: tuple[float, float] | None,
    display_landscape: bool,
) -> vp.Document:
    """Extract geometries from a SVG file.

    FILE may be a file path path or a dash (-) to read from the standard input instead.

    By default, the `read` command attempts to preserve the layer structure of the SVG. In this
    context, top-level groups (<g>) are each considered a layer. If any, all non-group,
    top-level SVG elements are imported into layer 1.

    The following logic is used to determine in which layer each SVG top-level group is
    imported:

        - If a `inkscape:label` attribute is present and contains digit characters, it is \
stripped of non-digit characters the resulting number is used as target layer. If the \
resulting number is 0, layer 1 is used instead.

        - If the previous step fails, the same logic is applied to the `id` attribute.

        - If both previous steps fail, the target layer matches the top-level group's order \
of appearance.

    Alternatively, geometries may be sorted into layers based on their attributes, such as
    color or stroke width. This is enabled by using the `--attr` option with the attribute
    to be considered. Multiple `--attr` options may be passed with different attributes. In
    this case, layers will be created for each unique combination of the provided attributes.

    Using `--single-layer`, the `read` command operates in single-layer mode. In this mode,
    all geometries are in a single layer regardless of the group structure. The current target
    layer is used default and can be specified with the `--layer` option. If the `--layer`
    option is used, `--single-layer` is assumed even if not explicitly provided.

    This command only extracts path elements as well as primitives (rectangles, ellipses,
    lines, polylines, polygons). Other elements such as text and bitmap images are discarded,
    and so is all formatting.

    All curved primitives (e.g. bezier path, ellipses, etc.) are linearized and approximated by
    polylines. The quantization length controls the maximum length of individual segments.

    Optionally, a line simplification with tolerance set to quantization can be applied on the
    SVG's curved element (e.g. circles, ellipses, arcs, bezier curves, etc.). This is enabled
    with the `--simplify` flag. This process reduces significantly the number of segments used
    to approximate the curve while still guaranteeing an accurate conversion, but may increase
    the execution time of this command.

    The `--parallel` option enables multiprocessing for the SVG conversion. This is recommended
    ONLY when using `--simplify` on large SVG files with many curved elements.

    By default, the geometries are cropped to the SVG boundaries defined by its width and
    length attributes. The crop operation can be disabled with the `--no-crop` option.

    In general, SVG boundaries are determined by the `width` and `height` of the top-level
    <svg> tag. However, the some SVG may have their width and/or height specified as percent
    value or even miss them altogether (in which case they are assumed to be set to 100%). In
    these cases, vpype attempts to use the `viewBox` attribute to determine the page size, or
    revert to a 1000x1000px default. The options `--display-size FORMAT` and
    `--display-landscape` can be used to specify a different format in such instances.

    When importing the SVG, the `read` commands attempts to extract the SVG attributes that
    are common to all paths within a layer. The "stroke", "stroke-width" and "inkscape:label"
    attributes are used to set the layer color, pen width and, respectively, name. The other
    attributes (e.g. "stroke-dasharray", etc.) are stored as layer properties with a `svg_`
    prefix. These properties are ignored by vpype but may be used by plug-ins. Also, the
    `write` command can optionally restore them in the exported SVG.

    Examples:

        Multi-layer SVG import:

            vpype read input_file.svg [...]

        Import SVG, sorting geometries by stroke color:

            vpype read --attr stroke input_file.svg [...]

        Import SVG, sorting geometries by stroke color and width:

            vpype read --attr stroke --attr stroke-width input_file.svg [...]

        Single-layer import:

            vpype read --single-layer input_file.svg [...]

        Single-layer import with target layer:

            vpype read --single-layer --layer 3 input_file.svg [...]

        Multi-layer import with specified quantization and line simplification enabled:

            vpype read --quantization 0.01mm --simplify input_file.svg [...]

        Multi-layer import with cropping disabled:

            vpype read --no-crop input_file.svg [...]
    """

    default_width = default_height = None
    if display_size is not None:
        default_width, default_height = display_size
        if display_landscape and default_width < default_height:
            default_width, default_height = default_height, default_width

    if file == "-":
        file = sys.stdin
        path: pathlib.Path | None = None
    elif not (path := pathlib.Path(file)).is_file():
        if no_fail:
            logging.debug("read: file doesn't exist, ignoring due to `--no-fail`")
            return document
        else:
            raise click.BadParameter(f"file {file!r} does not exist")

    if layer is not None and not single_layer:
        single_layer = True
        logging.debug("read: `--layer` provided, assuming single-layer mode")

    if single_layer:
        if len(attr) > 0:
            logging.warning("read: `--attr` is ignored in single-layer mode")

        lc, width, height = vp.read_svg(
            file,
            quantization=quantization,
            crop=not no_crop,
            simplify=simplify,
            parallel=parallel,
            default_width=default_width,
            default_height=default_height,
        )

        document.add(lc, single_to_layer_id(layer, document), with_metadata=True)
        document.extend_page_size((width, height))

        # vp_source is not set here, as it becomes a layer property
        if path:
            document.sources |= {path.absolute()}
    else:
        if len(attr) == 0:
            doc = vp.read_multilayer_svg(
                file,
                quantization=quantization,
                crop=not no_crop,
                simplify=simplify,
                parallel=parallel,
                default_width=default_width,
                default_height=default_height,
            )
        else:
            doc = vp.read_svg_by_attributes(
                file,
                attributes=attr,
                quantization=quantization,
                crop=not no_crop,
                simplify=simplify,
                parallel=parallel,
                default_width=default_width,
                default_height=default_height,
            )
        document.extend(doc)

    return document



from __future__ import annotations

import importlib.util

import click

import vpype as vp

from .cli import cli
from .decorators import generator

__all__ = ("script",)


@cli.command(group="Input")
@click.argument("file", type=click.Path(exists=True, dir_okay=False))
@generator
def script(file) -> vp.LineCollection:
    """
    Call an external python script to generate geometries.

    The script must contain a `generate()` function which will be called without arguments. It
    must return the generated geometries in one of the following format:

        - Shapely's MultiLineString
        - Iterable of Nx2 numpy float array
        - Iterable of Nx1 numpy complex array (where the real and imag part corresponds to
          the x, resp. y coordinates)

    All coordinates are expected to be in SVG pixel units (1/96th of an inch).
    """

    try:
        spec = importlib.util.spec_from_file_location("<external>", file)
        if spec is None:
            raise FileNotFoundError(f"file {file} not found")
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore
        return LineCollection(module.generate())  # type: ignore
    except Exception as exc:
        raise click.ClickException(
            f"the file path must point to a Python script containing a `generate()`"
            f"function ({str(exc)})"
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Generator

import click

import vpype as vp

from .substitution import (
    ExpressionSubstitutionError,
    PropertySubstitutionError,
    SubstitutionHelper,
)

__all__ = ("State", "_DeferredEvaluator")


class _DeferredEvaluator(ABC):
    """Abstract base-class for deferred evaluation of command parameters.

    Click parses all command arguments and options before running the corresponding command
    functions. In contrast, :ref:`property substitution <fundamentals_property_substitution>`
    must be performed just *just before* a command is executed, since it depends on the current
    context (e.g. currently defined properties and current layer ID).

    To address this issue, sub-classes of :class:`click.ParamType` may return some
    :class:`_DeferredEvaluator` instance instead of directly converting the user
    input. The command decorators (e.g. :func:`generator` and friends) will detect such
    instances, perform the conversion, and forward the converted value to the command function.
    """

    def __init__(self, text: str, param_name: str, *args, **kwargs):
        self._text = text
        self._param_name = param_name

    @abstractmethod
    def evaluate(self, state: State) -> Any:
        """Sub-class must override this function and return the converted value of
        ``self._text``"""

    def __str__(self):
        return self._text

    def __repr__(self):
        return repr(self._text)


class State:
    """Encapsulates the current state of the *vpype* pipeline processing.

    This class encapsulates the current state of the pipeline and provides services to
    commands. To access the current state instance, a command must use the :func:`pass_state`
    decorator.

    Args:
        document: if provided, use this document
    """

    _current_state: State | None = None

    def __init__(self, document: vp.Document | None = None):
        #: Content of the current pipeline.
        self.document: vp.Document = document or vp.Document()

        #: Default layer ID used by :func:`generator` and :func:`layer_processor` commands
        #: when ``--layer`` is not provided.
        self.target_layer_id: int | None = 1

        #: Layer ID being populated by a :func:`generator` or processed by a
        #: :func:`layer_processor` command.
        self.current_layer_id: int | None = None

        self._interpreter = SubstitutionHelper(self)

    def preprocess_argument(self, arg: Any) -> Any:
        """Evaluate an argument.

        If ``arg`` is a :class:`_DeferredEvaluator` instance, evaluate it a return its value
        instead.

        Args:
            arg: argument to evaluate

        Returns:
            returns the fully evaluated ``arg``
        """
        if isinstance(arg, tuple):
            return tuple(self.preprocess_argument(item) for item in arg)
        else:
            try:
                return arg.evaluate(self) if isinstance(arg, _DeferredEvaluator) else arg
            except Exception as exc:
                raise click.BadParameter(str(exc)) from exc

    def preprocess_arguments(
        self, args: tuple[Any, ...], kwargs: dict[str, Any]
    ) -> tuple[tuple[Any, ...], dict[str, Any]]:
        """Evaluate any instance of :class:`_DeferredEvaluator` and replace them with the
        converted value.
        """
        return (
            tuple(self.preprocess_argument(arg) for arg in args),
            {k: self.preprocess_argument(v) for k, v in kwargs.items()},
        )

    def substitute(self, text: str) -> str:
        """Apply :ref:`property <fundamentals_property_substitution>` and
        :ref:`expression <fundamentals_expression_substitution>` substitution on user input.

        Args:
            text: user input on which to perform the substitution

        Returns:
            fully substituted text
        """
        try:
            return self._interpreter.substitute(text)
        except (ExpressionSubstitutionError, PropertySubstitutionError) as exc:
            cause_err = "Error"
            cause = getattr(exc, "__cause__", None)
            if cause:
                cause_err = cause.__class__.__name__
            type_err = (
                "expression" if isinstance(exc, ExpressionSubstitutionError) else "property"
            )
            details = ""
            if cause and len(cause.args) > 0:
                details = ": " + cause.args[0]
            raise click.BadParameter(
                f"{cause_err} with {type_err} substitution{details}"
            ) from exc

    @classmethod
    def get_current(cls):
        """Returns the current :class:`State` instance.

        Commands should use the :func:`pass_state` decorator instead of using this function.
        """
        return cls._current_state

    @contextmanager
    def current(self):
        """Context manager to set the current state (used internally)."""
        self.__class__._current_state = self
        yield
        self.__class__._current_state = None

    @contextmanager
    def expression_variables(self, variables: dict[str, Any]) -> Generator[None, None, None]:
        """Context manager to temporarily set expression variables.

        This context manager is typically used by block processors to temporarily set relevant
        expression variables. These variables are deleted or, if pre-existing, restored upon
        exiting the context.

        Args:
            variables: variables to set

        Example::

            >>> import vpype_cli
            >>> @vpype_cli.cli.command()
            ... @vpype_cli.block_processor
            ... def run_twice(state, processors):
            ...     with state.expression_variables({"_first": True}):
            ...         vpype_cli.execute_processors(processors, state)
            ...     with state.expression_variables({"_first": False}):
            ...         vpype_cli.execute_processors(processors, state)
        """

        symtable = self._interpreter.symtable
        saved_items = {k: symtable[k] for k in variables if k in symtable}
        symtable.update(variables)
        yield
        for name in variables:
            symtable.pop(name)
        symtable.update(saved_items)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vpype',
 'vpype_cli',
 'vpype_viewer',
 'vpype_viewer.image',
 'vpype_viewer.qtviewer']

package_data = \
{'': ['*'],
 'vpype': ['fonts/*'],
 'vpype_viewer': ['resources/*', 'shaders/*'],
 'vpype_viewer.qtviewer': ['resources/*']}

install_requires = \
['Shapely[vectorized]>=1.8.2',
 'asteval>=0.9.26',
 'cachetools>=4.2.2',
 'click>=8.0.1,<8.2.0',
 'multiprocess>=0.70.11',
 'numpy>=1.20',
 'pnoise>=0.2.0',
 'scipy>=1.6',
 'svgelements>=1.6.10',
 'svgwrite>=1.4,<1.5',
 'tomli>=2.0.0']

extras_require = \
{'all': ['matplotlib>=3.3.2',
         'glcontext>=2.3.2',
         'moderngl>=5.6.2',
         'Pillow>=9.0.1',
         'PySide6>=6.3.2,<6.4.0']}

entry_points = \
{'console_scripts': ['vpype = vpype_cli.cli:cli']}

setup_kwargs = {
    'name': 'vpype',
    'version': '1.12.0',
    'description': 'The Swiss Army knife of vector graphics for pen plotters',
    'long_description': '![banner](https://github.com/abey79/vpype/raw/master/docs/images/banner_ua.png)\n\n\n# _vpype_\n\n[![PyPI](https://img.shields.io/pypi/v/vpype?label=PyPI&logo=pypi)](https://pypi.org/project/vpype/)\n[![python](https://img.shields.io/github/languages/top/abey79/vpype)](https://www.python.org)\n[![Downloads](https://pepy.tech/badge/vpype)](https://pepy.tech/project/vpype)\n[![license](https://img.shields.io/pypi/l/vpype)](https://vpype.readthedocs.io/en/latest/license.html)\n![Test](https://img.shields.io/github/workflow/status/abey79/vpype/Lint%20and%20Tests?label=Tests&logo=github)\n[![codecov](https://codecov.io/gh/abey79/vpype/branch/master/graph/badge.svg?token=CE7FD9D6XO)](https://codecov.io/gh/abey79/vpype)\n[![Sonarcloud Status](https://sonarcloud.io/api/project_badges/measure?project=abey79_vpype&metric=alert_status)](https://sonarcloud.io/dashboard?id=abey79_vpype)\n[![Documentation Status](https://img.shields.io/readthedocs/vpype?label=Read%20the%20Docs&logo=read-the-docs)](https://vpype.readthedocs.io/en/latest/?badge=latest)\n[![Code style: Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n_vpype_ is the Swiss-Army-knife command-line tool for plotter vector graphics.\n\n#### Contents\n\n* [What _vpype_ is?](#what-vpype-is)\n* [How does it work?](#how-does-it-work)\n* [Examples](#examples)\n* [What _vpype_ isn\'t?](#what-vpype-isnt)\n* [Installation](#installation)\n* [Documentation](#documentation)\n* [Feature overview](#feature-overview)\n    * [General](#general)\n    * [Input/Output](#inputoutput)\n    * [Layout and transforms](#layout-and-transforms)\n    * [Plotting optimization](#plotting-optimization)\n    * [Generation](#generation)\n    * [Extensibility and API](#extensibility-and-api)\n* [Plug-ins](#plug-ins)\n* [Contributing](#contributing)\n* [License](#license)\n\n\n## What _vpype_ is?\n\n_vpype_ is the Swiss-Army-knife command-line tool for plotter vector graphics. Here is what it can do:\n - **layout** existing vector files with precise control on position, scale and page format;\n - **optimize** existing SVG files for faster and cleaner plots;\n - create **HPGL output** for vintage plotters;\n - create **generative artwork** from scratch through built-in commands or [plug-ins](#plug-ins);\n - create, modify and process **multi-layer vector files** for multi-colour plots;\n - and much more...\n \n_vpype_ is highly **extensible** through plug-ins that can greatly extend its capabilities. For example, plug-ins\nalready exists for plotting [pixel art](https://github.com/abey79/vpype-pixelart),\n[half-toning with hatches](https://github.com/abey79/hatched), and much more. See below for a\n[list of existing plug-ins](#plug-ins).\n\n_vpype_ is also a [well documented](https://vpype.readthedocs.io/en/latest/api.html) **Python library**\nuseful to create generative art and tools for plotters. It includes data structures, utility and I/O functions, as well as\na hardware-accelerated flexible viewer for vector graphics. For example, the plotter generative art environment\n[vsketch](https://github.com/abey79/vsketch) is built upon _vpype_.\n\nCheck the [documentation](https://vpype.readthedocs.io/en/latest/) for a more thorough introduction to _vpype_.\n\n## How does it work?\n\n_vpype_ works by building so-called _pipelines_ of _commands_, where each command\'s output is fed to the next command\'s input.\nSome commands load geometries into the pipeline (e.g. the [`read`](https://vpype.readthedocs.io/en/latest/reference.html#read)\ncommand which loads geometries from a SVG file). Other commands modify these geometries, e.g. by cropping\nthem ([`crop`](https://vpype.readthedocs.io/en/latest/reference.html#crop)) or reordering them to minimize pen-up\ntravels ([`linesort`](https://vpype.readthedocs.io/en/latest/reference.html#linesort)). Finally, some other commands\njust read the geometries in the pipeline for display purposes ([`show`](https://vpype.readthedocs.io/en/latest/reference.html#show))\nor output to file ([`write`](https://vpype.readthedocs.io/en/latest/reference.html#write)).\n\nPipeline are defined using the _vpype_\'s CLI (command-line interface) in a terminal by typing `vpype` followed by the\nlist of commands, each with their optional parameters and their arguments:\n\n![command line](https://github.com/abey79/vpype/raw/master/docs/images/command_line.svg)\n\nThis pipeline uses five commands (in bold):\n- [`read`](https://vpype.readthedocs.io/en/latest/reference.html#read) loads geometries from a SVG file.\n- [`linemerge`](https://vpype.readthedocs.io/en/latest/reference.html#linemerge) merges paths whose extremities are close to each other (within the provided tolerance).\n- [`linesort`](https://vpype.readthedocs.io/en/latest/reference.html#linesort) reorder paths such as to minimise the pen-up travel.\n- [`crop`](https://vpype.readthedocs.io/en/latest/reference.html#crop), well, crops.\n- [`write`](https://vpype.readthedocs.io/en/latest/reference.html#write) export the resulting geometries to a SVG file.\n\nThere are many more commands available in *vpype*, see the [overview](#feature-overview) below.\n\nSome commands have arguments, which are always required (in italic). For example, a file path must be provided to the\n[`read`](https://vpype.readthedocs.io/en/latest/reference.html#read) command and dimensions must be provided to the [`crop`](https://vpype.readthedocs.io/en/latest/reference.html#crop) commands. A command may also have options which are, well,\noptional. In this example, `--page-size a4` means that the [`write`](https://vpype.readthedocs.io/en/latest/reference.html#write) command will generate a A4-sized SVG (otherwise it\nwould have the same size as _in.svg_). Likewise, because `--center` is used, the [`write`](https://vpype.readthedocs.io/en/latest/reference.html#write) command will center geometries\non the page before saving the SVG (otherwise the geometries would have been left at their original location).\n\n\n## Examples\n\n**Note**: The following examples are laid out over multiple lines using end-of-line escaping (`\\`). This is done to highlight the various commands of which the pipeline is made and would typically not be done in real-world use. \n\nLoad an SVG file, scale it to a specific size, and export it centered on an A4-sized, ready-to-plot SVG file:\n```bash\n$ vpype \\\n  read input.svg \\\n  layout --fit-to-margins 2cm a4 \\\n  write output.svg\n```\n\nOptimize paths to reduce plotting time (merge connected lines, sort them to minimize pen-up distance, randomize closed paths\' seam, and reduce the number of nodes):\n```bash\n$ vpype \\\n  read input.svg \\\n  linemerge --tolerance 0.1mm \\\n  linesort \\\n  reloop \\\n  linesimplify \\\n  write output.svg\n```\n\nLoad a SVG and display it in *vpype*\'s viewer, which enable close inspection of the layer and path structure):\n```bash\n$ vpype \\\n  read input.svg \\\n  show\n```\n\nLoad several SVG files and save them as a single, multi-layer SVG file (e.g. for multicolored drawings):\n```bash\n$ vpype \\\n  forfile "*.svg" \\\n    read --layer %_i% %_path% \\\n  end \\\n  write output.svg\n```\n\nExport a SVG to HPGL for vintage plotters:\n```bash\n$ vpype \\\n  read input.svg \\\n  layout --fit-to-margins 2cm --landscape a4 \\\n  write --device hp7475a output.hpgl\n```\n\nDraw the layer name on a SVG (this example uses [property substitution](https://vpype.readthedocs.io/en/latest/fundamentals.html#property-substitution)):\n```bash\n$ vpype \\\n    read input.svg \\\n    text --layer 1 "{vp_name}" \\\n    write output.svg\n```\n\nMerge multiple SVG files in a grid layout (this example uses [expression substitution](https://vpype.readthedocs.io/en/latest/fundamentals.html#expression-substitution)):\n```bash\n$ vpype \\\n    eval "files=glob(\'*.svg\')" \\\n    eval "cols=3; rows=ceil(len(files)/cols)" \\\n    grid -o 10cm 10cm "%cols%" "%rows%" \\\n        read --no-fail "%files[_i] if _i < len(files) else \'\'%" \\\n        layout -m 0.5cm 10x10cm \\\n    end \\\n    write combined_on_a_grid.svg\n```\n\nAn interactive version of the previous example is available in `examples/grid.vpy`. It makes use of `input()` expressions to ask parameters from the user:\n```bash\n$ vpype -I examples/grid.vpy\nFiles [*.svg]?\nNumber of columns [3]? 4\nColumn width [10cm]?\nRow height [10cm]? 15cm\nMargin [0.5cm]?\nOutput path [output.svg]?\n```\n\nSplit a SVG into one file per layer:\n```bash\n$ vpype \\\n    read input.svg \\\n    forlayer \\\n      write "output_%_name or _lid%.svg" \\\n    end\n```\n\nMore examples and recipes are available in the [cookbook](https://vpype.readthedocs.io/en/latest/cookbook.html). \n  \n## What _vpype_ isn\'t?\n\n_vpype_ caters to plotter generative art and does not aim to be a general purpose (think\nIllustrator/InkScape) vector graphic tools. One of the main reason for this is the fact _vpype_ converts everything \ncurvy (circles, bezier curves, etc.) to lines made of small segments. _vpype_ does import metadata such stroke and fill color, stroke width, etc., it only makes partial use of them and does not aim to maintain a full consistency with the SVG specification. These design choices make _vpype_\'s rich feature set possible, but limits its use for, e.g., printed media. \n \n \n## Installation\n\nDetailed installation instructions are available in the [latest documentation](https://vpype.readthedocs.io/en/latest/install.html).\n\nTL;DR:\n- Python 3.10 is recommended, but *vpype* is also compatible with Python 3.8 and 3.9. \n- *vpype* is published on the [Python Package Index](https://pypi.org) and can be installed using [pipx](https://pypa.github.io/pipx/):\n  ```bash\n  pipx install "vpype[all]"\n  ```\n- A Windows installer is available [here](https://github.com/abey79/vpype/releases), but plug-ins cannot be installed\nwhen using this method).\n- A CLI-only version of *vpype* can be installed using the following command:\n  ```bash\n  pipx install vpype\n  ```\n  This version does not include the [`show`](https://vpype.readthedocs.io/en/latest/reference.html#show) command but does not require some of the dependencies which are more difficult or impossible to install on some platforms (such as matplotlib, PySide6, and ModernGL).\n\n\n## Documentation\n\nThe _vpype_ CLI includes its own, detailed documentation:\n\n```bash\nvpype --help          # general help and command list\nvpype COMMAND --help  # help for a specific command\n``` \n\nIn addition, the [online documentation](https://vpype.readthedocs.io/en/latest/) provides extensive background\ninformation on the fundamentals behind _vpype_, a cookbook covering most common tasks, the _vpype_ API documentation,\nand much more.\n\n\n## Feature overview\n\n#### General\n\n- Easy to use **CLI** interface with integrated help (`vpype --help`and `vpype COMMAND --help`) and support for arbitrary units (e.g. `vpype read input.svg translate 3cm 2in`). \n- First-class **multi-layer support** with global or per-layer processing (e.g. `vpype COMMANDNAME --layer 1,3`) and optionally-probabilistic layer edition commands ([`lmove`](https://vpype.readthedocs.io/en/latest/reference.html#lmove), [`lcopy`](https://vpype.readthedocs.io/en/latest/reference.html#lcopy), [`ldelete`](https://vpype.readthedocs.io/en/latest/reference.html#ldelete), [`lswap`](https://vpype.readthedocs.io/en/latest/reference.html#lswap), [`lreverse`](https://vpype.readthedocs.io/en/latest/reference.html#lreverse)).\n- Support for **per-layer and global properties**, which acts as metadata and is used by multiple commands and plug-ins.\n- Support for [**property**](https://vpype.readthedocs.io/en/latest/fundamentals.html#property-substitution) and [**expression substitution**](https://vpype.readthedocs.io/en/latest/fundamentals.html#expression-substitution) in CLI user input.\n- Support for complex, **per-layer** processing ([`perlayer`](https://vpype.readthedocs.io/en/latest/reference.html#perlayer)).\n- Powerful hardware-accelerated **display** command with adjustable units, optional per-line coloring, optional pen-up trajectories display and per-layer visibility control ([`show`](https://vpype.readthedocs.io/en/latest/reference.html#show)).\n- Geometry **statistics** extraction ([`stat`](https://vpype.readthedocs.io/en/latest/reference.html#stat)).\n- Support for  **command history** recording (`vpype -H [...]`)\n- Support for **RNG seed** configuration for generative plug-ins (`vpype -s 37 [...]`).\n\n\n#### Input/Output\n\n- Single- and multi-layer **SVG input** with adjustable precision, parallel processing for large SVGs, and supports percent or missing width/height ([`read`](https://vpype.readthedocs.io/en/latest/reference.html#read)).\n- Support for **SVG output** with fine layout control (page size and orientation, centering), layer support with custom layer names, optional display of pen-up trajectories, various option for coloring ([`write`](https://vpype.readthedocs.io/en/latest/reference.html#write)).\n- Support for **HPGL output** config-based generation of HPGL code with fine layout control (page size and orientation, centering).\n- Support for pattern-based **file collection** processing ([`forfile`](https://vpype.readthedocs.io/en/latest/reference.html#forfile)). \n\n\n#### Layout and transforms\n\n- Easy and flexible **layout** command for centring and fitting to margin with selectable le horizontal and vertical alignment\n  ([`layout`](https://vpype.readthedocs.io/en/latest/reference.html#layout)).\n- **Page rotation** command ([`pagerotate`](https://vpype.readthedocs.io/en/latest/reference.html#pagerotate)).\n- Powerful **transform** commands for scaling, translating, skewing and rotating geometries ([`scale`](https://vpype.readthedocs.io/en/latest/reference.html#scale), [`translate`](https://vpype.readthedocs.io/en/latest/reference.html#translate), [`skew`](https://vpype.readthedocs.io/en/latest/reference.html#skew), [`rotate`](https://vpype.readthedocs.io/en/latest/reference.html#rotate)).\n- Support for **scaling** and **cropping** to arbitrary dimensions ([`scaleto`](https://vpype.readthedocs.io/en/latest/reference.html#scaleto), [`crop`](https://vpype.readthedocs.io/en/latest/reference.html#crop)).\n- Support for **trimming** geometries by an arbitrary amount ([`trim`](https://vpype.readthedocs.io/en/latest/reference.html#trim)).\n- Arbitrary **page size** definition ([`pagesize`](https://vpype.readthedocs.io/en/latest/reference.html#pagesize)). \n\n\n#### Metadata\n\n- Adjust layer **color**, **alpha**, **pen width** and **name** ([`color`](https://vpype.readthedocs.io/en/latest/reference.html#color), [`alpha`](https://vpype.readthedocs.io/en/latest/reference.html#alpha), [`penwidth`](https://vpype.readthedocs.io/en/latest/reference.html#penwidth), [`name`](https://vpype.readthedocs.io/en/latest/reference.html#name)).\n- Apply provided or fully customisable **pen configurations** ([`pens`](https://vpype.readthedocs.io/en/latest/reference.html#pens)).\n- Manipulate global and per-layer **properties** ([`propset`](https://vpype.readthedocs.io/en/latest/reference.html#propset), [`propget`](https://vpype.readthedocs.io/en/latest/reference.html#propget), [`proplist`](https://vpype.readthedocs.io/en/latest/reference.html#proplist), [`propdel`](https://vpype.readthedocs.io/en/latest/reference.html#propdel), [`propclear`](https://vpype.readthedocs.io/en/latest/reference.html#propclear)).\n\n\n#### Plotting optimization\n\n- **Line merging** with optional path reversal and configurable merging threshold ([`linemerge`](https://vpype.readthedocs.io/en/latest/reference.html#linemerge)).\n- **Line sorting** with optional path reversal ([`linesort`](https://vpype.readthedocs.io/en/latest/reference.html#linesort)).\n- **Line simplification** with adjustable accuracy ([`linesimplify`](https://vpype.readthedocs.io/en/latest/reference.html#linesimplify)).\n- Closed paths\' **seam location randomization**, to reduce the visibility of pen-up/pen-down artifacts ([`reloop`](https://vpype.readthedocs.io/en/latest/reference.html#reloop)).\n- Support for generating **multiple passes** on each line ([`multipass`](https://vpype.readthedocs.io/en/latest/reference.html#multipass)).\n\n#### Filters\n\n- Support for **filtering** by line lengths or closed-ness ([`filter`](https://vpype.readthedocs.io/en/latest/reference.html#filter)).\n- **Squiggle** filter for shaky-hand or liquid-like styling ([`squiggles`](https://vpype.readthedocs.io/en/latest/reference.html#squiggles))\n- Support for **splitting** all lines to their constituent segments ([`splitall`](https://vpype.readthedocs.io/en/latest/reference.html#splitall)).\n- Support for **reversing** order of paths within their layers ([`reverse`](https://vpype.readthedocs.io/en/latest/reference.html#reverse)).\n- Support for **splitting** layers by drawing distance ([`splitdist`](https://vpype.readthedocs.io/en/latest/reference.html#splitdist))\n\n#### Generation\n \n - Generation of arbitrary **primitives** including lines, rectangles, circles, ellipses and arcs ([`line`](https://vpype.readthedocs.io/en/latest/reference.html#line), [`rect`](https://vpype.readthedocs.io/en/latest/reference.html#rect), [`circle`](https://vpype.readthedocs.io/en/latest/reference.html#circle), [`ellipse`](https://vpype.readthedocs.io/en/latest/reference.html#ellipse), [`arc`](https://vpype.readthedocs.io/en/latest/reference.html#arc)).\n - Generation of **text** using bundled Hershey fonts ([`text`](https://vpype.readthedocs.io/en/latest/reference.html#text))\n - Generation of grid-like layouts ([`grid`](https://vpype.readthedocs.io/en/latest/reference.html#grid)).\n - Generation of a **frame** around the geometries ([`frame`](https://vpype.readthedocs.io/en/latest/reference.html#frame)).\n - Generation of random lines for debug/learning purposes ([`random`](https://vpype.readthedocs.io/en/latest/reference.html#random))\n\n#### Extensibility and API\n\n - First-class support for **plug-in** extensions (e.g [vpype-text](https://github.com/abey79/vpype-text), [hatched](https://github.com/abey79/hatched), [occult](https://github.com/LoicGoulefert/occult)).\n - Support for **script-based** generation ([`script`](https://vpype.readthedocs.io/en/latest/reference.html#script)).\n - Powerful and [well-documented](https://vpype.readthedocs.io/en/latest/api.html) **API** for plug-ins and other plotter generative art projects.\n \n \n ## Plug-ins\n \n Here is a list of known vpype plug-ins (please make a pull request if yours is missing):\n \n - [vsketch](https://github.com/abey79/vsketch): *vsketch* is complete framework for plotter generative artists implemented using *vpype*\'s API\n - [vpype-perspective](https://github.com/abey79/vpype-perspective): put your art in perspective\n - [vpype-pixelart](https://github.com/abey79/vpype-pixelart): plot pixel art\n - [hatched](https://github.com/abey79/hatched): half-toning with hatches\n - [vpype-flow-imager](https://github.com/serycjon/vpype-flow-imager): convert images to flow-line-based designs\n - [occult](https://github.com/LoicGoulefert/occult): perform hidden line removal\n - [deduplicate](https://github.com/LoicGoulefert/deduplicate): remove duplicate lines\n - [vpype-explorations](https://github.com/abey79/vpype-explorations): my personal grab bag of experiments and utilities\n - [vpype-gcode](https://github.com/tatarize/vpype-gcode/): flexible export command for gcode and other text-based format\n - [vpype-dxf](https://github.com/tatarize/vpype-dxf/): read dxf files\n - [vpype-embroidery](https://github.com/EmbroidePy/vpype-embroidery): various embroidery-related utilities, including read from/write to most embroidery formats \n - [vpype-vectrace](https://github.com/tatarize/vpype-vectrace): create outlines from images with vector tracing\n \n \n ## Contributing\n\nContributions to this project are welcome and do not necessarily require software development skills! Check the\n[Contributing section](https://vpype.readthedocs.io/en/latest/contributing.html) of the documentation for more\ninformation.  \n\n\n## License\n\nThis project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details.\n',
    'author': 'Antoine Beyeler',
    'author_email': 'abeyeler@ab-ware.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/abey79/vpype',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
Qt Viewer
"""
from __future__ import annotations

import functools
import logging
import math
import os
import sys

import moderngl as mgl
from PySide6.QtCore import QEvent, QSettings, QSize, Qt, Signal
from PySide6.QtGui import (
    QAction,
    QActionGroup,
    QMouseEvent,
    QNativeGestureEvent,
    QScreen,
    QSurfaceFormat,
    QWheelEvent,
)
from PySide6.QtOpenGLWidgets import QOpenGLWidget
from PySide6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QMenu,
    QSizePolicy,
    QToolBar,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

import vpype as vp

from .._scales import UnitType
from ..engine import Engine, ViewMode
from .utils import PenOpacityActionGroup, PenWidthActionGroup, load_icon, set_sigint_handler

__all__ = ["QtViewerWidget", "QtViewer", "show"]


_DEBUG_ENABLED = "VPYPE_VIEWER_DEBUG" in os.environ


# handle UI scaling
def _configure_ui_scaling():
    viewer_config = vp.config_manager.config.get("viewer", {})
    if "QT_SCALE_FACTOR" not in os.environ and "ui_scale_factor" in viewer_config:
        os.environ["QT_SCALE_FACTOR"] = str(viewer_config["ui_scale_factor"])


_configure_ui_scaling()

# set default format
default_format = QSurfaceFormat()
default_format.setVersion(3, 3)
default_format.setProfile(QSurfaceFormat.CoreProfile)
default_format.setSamples(4)
QSurfaceFormat.setDefaultFormat(default_format)


class QtViewerWidget(QOpenGLWidget):
    """QGLWidget wrapper around :class:`Engine` to display a :class:`vpype.Document` in
    Qt GUI."""

    mouse_coords = Signal(str)

    def __init__(self, document: vp.Document | None = None, parent=None):
        """Constructor.
        Args:
            document: the document to display
            parent: QWidget parent
        """

        super().__init__(parent=parent)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.setMouseTracking(True)

        self._document = document

        self._last_mouse_x = 0.0
        self._last_mouse_y = 0.0
        self._mouse_drag = False
        self._factor = 1.0

        # deferred initialization in initializeGL()
        self._ctx: mgl.Context | None = None
        self._framebuffer = None
        self._inited = False
        self.engine = Engine(
            view_mode=ViewMode.OUTLINE, show_pen_up=False, render_cb=self.update
        )

        # print diagnostic information
        screen = self.screen()
        logging.info(
            f"QScreen info: pixelRatio={screen.devicePixelRatio()}, "
            f"physicalSize={screen.physicalSize().toTuple()}, "
            f"physicalDotsPerInch={screen.physicalDotsPerInch()}, "
            f"virtualSize={screen.virtualSize().toTuple()}, "
            f"size={screen.size().toTuple()}, "
            f"logicalDotsPerInch={screen.logicalDotsPerInch()}, "
        )

    def document(self) -> vp.Document | None:
        """Return the :class:`vpype.Document` currently assigned to the widget."""
        return self._document

    def set_document(self, document: vp.Document | None) -> None:
        """Assign a new :class:`vpype.Document` to the widget."""
        self._document = document
        self.engine.document = document

    def on_screen_changed(self, screen: QScreen):
        self._factor = screen.devicePixelRatio()
        self.engine.pixel_factor = self._factor

        # force an update and reset of viewport's dimensions
        self.resizeGL(round(self.geometry().width()), round(self.geometry().height()))

    def _initializeGL(self):
        """Initialize ModernGL context.

        This used to be an override of QGLWidget.initializeGL() with PySide2. However, as per
        PySide6 documentation, "the framebuffer is not yet available at this stage". We thus
        defer the initialization to the first call of QOpenGLWidget.paintGL().
        """

        # this fails in __init__() (windowHandle() returns None)
        self.window().windowHandle().screenChanged.connect(self.on_screen_changed)

        self._ctx = mgl.create_context()
        logging.info(f"Context info: {self._ctx.info}")

        self._framebuffer = self._ctx.detect_framebuffer()
        self._framebuffer.viewport = (
            0,
            0,
            int(self._factor * self.width()),
            int(self._factor * self.height()),
        )

        self.engine.post_init(
            self._ctx, int(self._factor * self.width()), int(self._factor * self.height())
        )
        self.on_screen_changed(self.screen())
        self.engine.document = self._document
        self.engine.fit_to_viewport()

    def paintGL(self):
        if not self._inited:
            self._initializeGL()
            self._inited = True

        self._framebuffer.use()
        self.engine.render()

    def resizeGL(self, w: int, h: int) -> None:
        width = int(w * self._factor)
        height = int(h * self._factor)
        self.engine.resize(width, height)

        if self._framebuffer:
            self._framebuffer.viewport = (0, 0, width, height)

    def mousePressEvent(self, evt: QMouseEvent):
        pos = evt.position()
        self._last_mouse_x = pos.x()
        self._last_mouse_y = pos.y()
        self._mouse_drag = True

    def mouseMoveEvent(self, evt: QMouseEvent):
        pos = evt.position()

        if self._mouse_drag:
            self.engine.pan(
                self._factor * (pos.x() - self._last_mouse_x),
                self._factor * (pos.y() - self._last_mouse_y),
            )
            self._last_mouse_x = pos.x()
            self._last_mouse_y = pos.y()

        # update mouse coordinate display
        if pos.x() < 0 or pos.x() > self.width() or pos.y() < 0 or pos.y() > self.height():
            # noinspection PyUnresolvedReferences
            self.mouse_coords.emit("")
        else:
            x, y = self.engine.viewport_to_model(
                self._factor * pos.x(), self._factor * pos.y()
            )
            spec = self.engine.scale_spec
            decimals = max(0, math.ceil(-math.log10(1 / spec.to_px / self.engine.scale)))
            # noinspection PyUnresolvedReferences
            self.mouse_coords.emit(
                f"{x / spec.to_px:.{decimals}f}{spec.unit}, "
                f"{y / spec.to_px:.{decimals}f}{spec.unit}"
            )

    def mouseReleaseEvent(self, evt):
        self._mouse_drag = False

    def leaveEvent(self, event: QEvent) -> None:
        # noinspection PyUnresolvedReferences
        self.mouse_coords.emit("")  # type: ignore

    def wheelEvent(self, event: QWheelEvent) -> None:
        if event.source() == Qt.MouseEventSource.MouseEventSynthesizedBySystem:
            # track pad
            scroll_delta = event.pixelDelta()
            self.engine.pan(self._factor * scroll_delta.x(), self._factor * scroll_delta.y())
        else:
            # mouse wheel
            zoom_delta = event.angleDelta().y()
            pos = event.position()
            self.engine.zoom(
                zoom_delta / 500.0, self._factor * pos.x(), self._factor * pos.y()
            )

    def event(self, event: QEvent) -> bool:
        # handle pinch zoom on mac
        if (
            isinstance(event, QNativeGestureEvent)
            and event.gestureType() == Qt.NativeGestureType.ZoomNativeGesture
        ):
            self.engine.zoom(
                2.0 * event.value(),
                event.localPos().x() * self._factor,
                event.localPos().y() * self._factor,
            )
            return True

        return super().event(event)


class QtViewer(QWidget):
    """Full featured, stand-alone viewer suitable for displaying a :class:`vpype.Document` to
    a user."""

    def __init__(
        self,
        document: vp.Document | None = None,
        view_mode: ViewMode = ViewMode.PREVIEW,
        show_pen_up: bool = False,
        show_points: bool = False,
        parent=None,
    ):
        super().__init__(parent)

        self._settings = QSettings()
        self._settings.beginGroup("viewer")

        self.setWindowTitle("vpype viewer")
        self.setStyleSheet(
            """
        QToolButton:pressed {
            background-color: rgba(0, 0, 0, 0.2);
        }
        """
        )

        self._viewer_widget = QtViewerWidget(parent=self)

        # setup toolbar
        self._toolbar = QToolBar()
        self._icon_size = QSize(32, 32)
        self._toolbar.setIconSize(self._icon_size)

        view_mode_grp = QActionGroup(self._toolbar)
        if _DEBUG_ENABLED:
            act = view_mode_grp.addAction("None")
            act.setCheckable(True)
            act.setChecked(view_mode == ViewMode.NONE)
            act.triggered.connect(  # type:ignore
                functools.partial(self.set_view_mode, ViewMode.NONE)
            )
        act = view_mode_grp.addAction("Outline Mode")
        act.setCheckable(True)
        act.setChecked(view_mode == ViewMode.OUTLINE)
        act.triggered.connect(  # type:ignore
            functools.partial(self.set_view_mode, ViewMode.OUTLINE)
        )
        act = view_mode_grp.addAction("Outline Mode (Colorful)")
        act.setCheckable(True)
        act.setChecked(view_mode == ViewMode.OUTLINE_COLORFUL)
        act.triggered.connect(  # type:ignore
            functools.partial(self.set_view_mode, ViewMode.OUTLINE_COLORFUL)
        )
        act = view_mode_grp.addAction("Preview Mode")
        act.setCheckable(True)
        act.setChecked(view_mode == ViewMode.PREVIEW)
        act.triggered.connect(  # type:ignore
            functools.partial(self.set_view_mode, ViewMode.PREVIEW)
        )
        self.set_view_mode(view_mode)

        # VIEW MODE
        # view modes
        view_mode_btn = QToolButton()
        view_mode_menu = QMenu(view_mode_btn)
        act = view_mode_menu.addAction("View Mode:")
        act.setEnabled(False)
        view_mode_menu.addActions(view_mode_grp.actions())
        view_mode_menu.addSeparator()
        # show pen up
        act = view_mode_menu.addAction("Show Pen-Up Trajectories")
        act.setCheckable(True)
        act.setChecked(show_pen_up)
        act.toggled.connect(self.set_show_pen_up)  # type:ignore
        self._viewer_widget.engine.show_pen_up = show_pen_up
        # show points
        act = view_mode_menu.addAction("Show Points")
        act.setCheckable(True)
        act.setChecked(show_points)
        act.toggled.connect(self.set_show_points)  # type:ignore
        self._viewer_widget.engine.show_points = show_points
        # preview mode options
        view_mode_menu.addSeparator()
        act = view_mode_menu.addAction("Preview Mode Options:")
        act.setEnabled(False)
        # pen width
        pen_width_menu = view_mode_menu.addMenu("Default Pen Width")
        act = pen_width_menu.addAction("Override")
        act.setCheckable(True)
        act.toggled.connect(self.set_override_pen_width)  # type:ignore
        pen_width_menu.addSeparator()
        act_grp = PenWidthActionGroup(0.3, parent=pen_width_menu)
        act_grp.triggered.connect(self.set_default_pen_width_mm)  # type:ignore
        pen_width_menu.addActions(act_grp.actions())
        self.set_default_pen_width_mm(0.3)

        # pen opacity
        pen_opacity_menu = view_mode_menu.addMenu("Default Pen Opacity")
        act = pen_opacity_menu.addAction("Override")
        act.setCheckable(True)
        act.toggled.connect(self.set_override_pen_opacity)  # type:ignore
        pen_opacity_menu.addSeparator()
        act_grp = PenOpacityActionGroup(0.8, parent=pen_opacity_menu)
        act_grp.triggered.connect(self.set_default_pen_opacity)  # type:ignore
        pen_opacity_menu.addActions(act_grp.actions())
        self.set_default_pen_opacity(0.8)

        # debug view
        if _DEBUG_ENABLED:
            act = view_mode_menu.addAction("Debug View")
            act.setCheckable(True)
            act.toggled.connect(self.set_debug)  # type:ignore
        # rulers
        view_mode_menu.addSeparator()
        act = view_mode_menu.addAction("Show Rulers")
        act.setCheckable(True)
        val = bool(self._settings.value("show_rulers", True))
        act.setChecked(val)
        act.toggled.connect(self.set_show_rulers)  # type:ignore
        self._viewer_widget.engine.show_rulers = val
        # units
        units_menu = view_mode_menu.addMenu("Units")
        unit_action_grp = QActionGroup(units_menu)
        unit_type = UnitType(self._settings.value("unit_type", UnitType.METRIC))
        act = unit_action_grp.addAction("Metric")
        act.setCheckable(True)
        act.setChecked(unit_type == UnitType.METRIC)
        act.setData(UnitType.METRIC)
        act = unit_action_grp.addAction("Imperial")
        act.setCheckable(True)
        act.setChecked(unit_type == UnitType.IMPERIAL)
        act.setData(UnitType.IMPERIAL)
        act = unit_action_grp.addAction("Pixel")
        act.setCheckable(True)
        act.setChecked(unit_type == UnitType.PIXELS)
        act.setData(UnitType.PIXELS)
        unit_action_grp.triggered.connect(self.set_unit_type)  # type:ignore
        units_menu.addActions(unit_action_grp.actions())
        self._viewer_widget.engine.unit_type = unit_type

        view_mode_btn.setMenu(view_mode_menu)
        view_mode_btn.setIcon(load_icon("eye-outline.svg"))
        view_mode_btn.setText("View")
        view_mode_btn.setPopupMode(QToolButton.InstantPopup)
        view_mode_btn.setStyleSheet("QToolButton::menu-indicator { image: none; }")
        self._toolbar.addWidget(view_mode_btn)

        # LAYER VISIBILITY
        self._layer_visibility_btn = QToolButton()
        self._layer_visibility_btn.setIcon(load_icon("layers-triple-outline.svg"))
        self._layer_visibility_btn.setText("Layer")
        self._layer_visibility_btn.setMenu(QMenu(self._layer_visibility_btn))
        self._layer_visibility_btn.setPopupMode(QToolButton.InstantPopup)
        self._layer_visibility_btn.setStyleSheet(
            "QToolButton::menu-indicator { image: none; }"
        )
        self._toolbar.addWidget(self._layer_visibility_btn)

        # FIT TO PAGE
        fit_act = self._toolbar.addAction(load_icon("fit-to-page-outline.svg"), "Fit")
        fit_act.triggered.connect(self._viewer_widget.engine.fit_to_viewport)  # type:ignore

        # RULER
        # TODO: not implemented yet
        # self._toolbar.addAction(load_icon("ruler-square.svg"), "Units")

        # MOUSE COORDINATES
        self._mouse_coord_lbl = QLabel("")
        self._mouse_coord_lbl.setContentsMargins(6, 6, 6, 6)
        self._mouse_coord_lbl.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self._mouse_coord_lbl.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self._toolbar.addWidget(self._mouse_coord_lbl)
        # noinspection PyUnresolvedReferences
        self._viewer_widget.mouse_coords.connect(self.set_mouse_coords)  # type: ignore

        # setup horizontal layout for optional side widgets
        self._hlayout = QHBoxLayout()
        self._hlayout.setSpacing(0)
        self._hlayout.setContentsMargins(0, 0, 0, 0)
        self._hlayout.addWidget(self._viewer_widget)
        widget = QWidget()
        widget.setLayout(self._hlayout)

        # setup global vertical layout
        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._toolbar)
        layout.addWidget(widget)
        self.setLayout(layout)

        if document is not None:
            self.set_document(document)

    def add_side_widget(self, widget: QWidget) -> None:
        self._hlayout.addWidget(widget)

    def set_document(self, document: vp.Document | None) -> None:
        self._viewer_widget.set_document(document)
        self._update_layer_menu()

    def _update_layer_menu(self):
        layer_menu = QMenu(self._layer_visibility_btn)
        doc = self._viewer_widget.document()
        if doc is not None:
            for layer_id in sorted(doc.layers):
                item_name = f"Layer {layer_id}"
                layer_name = doc.layers[layer_id].property(vp.METADATA_FIELD_NAME)
                if layer_name:
                    item_name += ": " + layer_name
                action = layer_menu.addAction(item_name)
                action.setCheckable(True)
                action.setChecked(True)
                # TODO: set color icon
                action.triggered.connect(
                    functools.partial(
                        self._viewer_widget.engine.toggle_layer_visibility, layer_id
                    )
                )
        self._layer_visibility_btn.setMenu(layer_menu)

    def set_mouse_coords(self, txt: str) -> None:
        self._mouse_coord_lbl.setText(txt)

    def set_view_mode(self, mode: ViewMode) -> None:
        self._viewer_widget.engine.view_mode = mode

    def set_show_pen_up(self, show_pen_up: bool) -> None:
        self._viewer_widget.engine.show_pen_up = show_pen_up

    def set_show_points(self, show_points: bool) -> None:
        self._viewer_widget.engine.show_points = show_points

    def set_default_pen_width_mm(self, value: float | QAction) -> None:
        if isinstance(value, QAction):
            value = float(value.data())
        self._viewer_widget.engine.default_pen_width = value / 25.4 * 96.0

    def set_default_pen_opacity(self, value: float | QAction) -> None:
        if isinstance(value, QAction):
            value = float(value.data())
        self._viewer_widget.engine.default_pen_opacity = value

    def set_override_pen_width(self, value: bool) -> None:
        self._viewer_widget.engine.override_pen_width = value

    def set_override_pen_opacity(self, value: bool) -> None:
        self._viewer_widget.engine.override_pen_opacity = value

    def set_debug(self, debug: bool) -> None:
        self._viewer_widget.engine.debug = debug

    def set_show_rulers(self, show_rulers: bool) -> None:
        self._viewer_widget.engine.show_rulers = show_rulers
        self._settings.setValue("show_rulers", show_rulers)

    def set_unit_type(self, sender: QAction) -> None:
        val = UnitType(sender.data())
        self._viewer_widget.engine.unit_type = val
        self._settings.setValue("unit_type", val)


def show(
    document: vp.Document,
    view_mode: ViewMode = ViewMode.PREVIEW,
    show_pen_up: bool = False,
    show_points: bool = False,
    argv=None,
) -> int:
    """Show a viewer for the provided :class:`vpype.Document` instance.

    This function returns when the user close the window.

    Args:
        document: the document to display
        view_mode: view mode
        show_pen_up: render pen-up trajectories
        show_points: render points
        argv: argument passed to Qt

    Returns:
        exit status returned by Qt
    """
    if argv is None and len(sys.argv) > 0:
        argv = [sys.argv[0]]

    app = QApplication.instance()
    if not app or not isinstance(app, QApplication):
        app = QApplication(argv)
        app.setOrganizationName("abey79")
        app.setOrganizationDomain("abey79.github.io")
        app.setApplicationName("vpype")
    app.setAttribute(Qt.AA_UseHighDpiPixmaps)

    widget = QtViewer(
        document, view_mode=view_mode, show_pen_up=show_pen_up, show_points=show_points
    )
    sz = app.primaryScreen().availableSize()
    widget.move(int(sz.width() * 0.05), int(sz.height() * 0.1))
    widget.resize(int(sz.width() * 0.9), int(sz.height() * 0.8))

    widget.show()

    # noinspection PyUnusedLocal
    def sigint_handler(signum, frame):
        QApplication.quit()

    with set_sigint_handler(sigint_handler):
        res = app.exec()

    return res



from __future__ import annotations

import random

import click

import vpype as vp

from .cli import cli
from .decorators import global_processor
from .types import LayerType, multiple_to_layer_ids, single_to_layer_id

__all__ = ("lcopy", "lmove", "ldelete", "lreverse", "lswap")


@cli.command(group="Layers")
@click.argument("sources", type=LayerType(accept_multiple=True))
@click.argument("dest", type=LayerType(accept_new=True))
@click.option(
    "-p",
    "--prob",
    type=click.FloatRange(0.0, 1.0),
    help="Path copy probability (default: 1.0).",
)
@click.option("-m", "--no-prop", is_flag=True, help="Do not copy metadata.")
@global_processor
def lcopy(document, sources, dest, prob: float | None, no_prop: bool):
    """Copy the content of one or more layer(s) to another layer.

    SOURCES can be a single layer ID, the string 'all' (to copy all non-empty layers,
    or a coma-separated, whitespace-free list of layer IDs.

    DEST can be a layer ID or the string 'new', in which case a new layer with the
    lowest available ID will be created.

    If a layer is both in the source and destination, its content is not duplicated.

    The `--prob` option controls the probability with which each path is copied. With a value
    lower than 1.0, some paths will not be copied to DEST, which may be used to achieve random
    coloring effects.

    If a single source layer is specified and the `--prob` option is not used, the properties
    of the source layer are copied to the destination layer, overwriting any existing
    properties with the same name. This behaviour can be disabled with the `--no-prop` option.

    Examples:

        Copy layer 1 to a new layer:

            vpype [...] lcopy 1 new [...]  # duplicate layer 1

        Make a new layer with a merged copy of layer 1 and 2:

            vpype [...] lcopy 1,2 new [...]  # make new layer with merged copy of layer 1 and 2

        Add a merged copy of all layers to layer 1. If layer 1 previously had content, this \
content is not duplicated:

            vpype [...] lcopy all 1 [...]
    """

    src_lids = multiple_to_layer_ids(sources, document)
    dest_lid = single_to_layer_id(dest, document)

    if dest_lid in src_lids:
        src_lids.remove(dest_lid)

    lc = vp.LineCollection()
    for lid in src_lids:
        if prob is not None:
            for line in document[lid]:
                if random.random() < prob:
                    lc.append(line)
        else:
            lc.extend(document[lid])

    if len(lc) > 0:
        document.add(lc, dest_lid)

    if len(src_lids) == 1 and prob is None and not no_prop:
        document.layers[dest_lid].metadata.update(document.layers[src_lids[0]].metadata)

    return document


@cli.command(group="Layers")
@click.argument("sources", type=LayerType(accept_multiple=True))
@click.argument("dest", type=LayerType(accept_new=True))
@click.option(
    "-p",
    "--prob",
    type=click.FloatRange(0.0, 1.0),
    help="Path move probability (default: 1.0).",
)
@click.option("-m", "--no-prop", is_flag=True, help="Do not move metadata.")
@global_processor
def lmove(document, sources, dest, prob: float | None, no_prop: bool):
    """Move the content of one or more layer(s) to another layer.

    SOURCES can be a single layer ID, the string 'all' (to copy all non-empty layers,
    or a coma-separated, whitespace-free list of layer IDs.

    DEST can be a layer ID or the string 'new', in which case a new layer with the
    lowest available ID will be created.

    Layer(s) left empty after the move are then discarded and may thus be reused by subsequent
    commands using 'new' as destination layer.

    The `--prob` option controls the probability with which each path is moved. With a value
    lower than 1.0, some paths will not be moved to DEST, which may be used to achieve random
    coloring effects.

    If a layer is both in the source and destination, its content is not duplicated.

    If a single source layer is specified and the `--prob` option is not used, the properties
    of the source layer are moved to the destination layer, overwriting any existing
    properties with the same name. This behaviour can be disabled with the `--no-prop` option.

    Examples:

        Merge layer 1 and 2 to layer 1 (the content of layer 1 is not duplicated):

            vpype [...] lmove 1,2 1 [...]  # merge layer 1 and 2 to layer 1
    """

    src_lids = multiple_to_layer_ids(sources, document)
    dest_lid = single_to_layer_id(dest, document)

    if dest_lid in document.layers:
        dest_lc = document.layers[dest_lid].clone()
    else:
        dest_lc = vp.LineCollection()

    move_metadata = len(src_lids) == 1 and prob is None and not no_prop
    source_metadata = document.layers[src_lids[0]].metadata if move_metadata else {}

    for lid in src_lids:
        if prob is not None:
            # split lines with provided probability
            remaining_lines = vp.LineCollection()
            moving_lines = vp.LineCollection()
            for line in document.layers[lid]:
                if random.random() < prob:
                    moving_lines.append(line)
                else:
                    remaining_lines.append(line)

            if len(remaining_lines) > 0:
                document.replace(remaining_lines, lid)
            else:
                document.pop(lid)

            if len(moving_lines) > 0:
                dest_lc.extend(moving_lines)
        else:
            dest_lc.extend(document.pop(lid))
            if move_metadata:
                dest_lc.metadata.update(source_metadata)

    if len(dest_lc) > 0:
        document.add(dest_lc, dest_lid, with_metadata=True)
    return document


@cli.command(group="Layers")
@click.argument("layers", type=LayerType(accept_multiple=True))
@click.option(
    "-k", "--keep", is_flag=True, help="Specified layers must be kept instead of deleted."
)
@click.option(
    "-p",
    "--prob",
    type=click.FloatRange(0.0, 1.0),
    help="Path deletion probability (default: 1.0).",
)
@global_processor
def ldelete(document: vp.Document, layers, keep: bool, prob: float | None) -> vp.Document:
    """Delete one or more layers.

    LAYERS can be a single layer ID, the string 'all' (to delete all layers), or a
    coma-separated, whitespace-free list of layer IDs.

    If the `--keep` option is used, the specified layers are kept and, instead, all other
    layers deleted.

    The `--prob` option controls the probability with which each path is deleted. With a value
    lower than 1.0, some paths will not be deleted.
    """

    lids = set(multiple_to_layer_ids(layers, document))

    if keep:
        lids = document.layers.keys() - lids

    for lid in lids:
        if prob is not None:
            lc = document.layers[lid].clone()
            for line in document[lid]:
                if not random.random() < prob:
                    lc.append(line)

            if len(lc) == 0:
                document.pop(lid)
            else:
                document[lid] = lc
        else:
            document.pop(lid)

    return document


@cli.command(group="Layers")
@click.argument("first", type=LayerType(accept_multiple=False, accept_new=False))
@click.argument("second", type=LayerType(accept_multiple=False, accept_new=False))
@click.option(
    "-p",
    "--prob",
    type=click.FloatRange(0.0, 1.0),
    help="Path deletion probability (default: 1.0).",
)
@click.option("-m", "--no-prop", is_flag=True, help="Do not move metadata.")
@global_processor
def lswap(
    document: vp.Document, first: int, second: int, prob: float | None, no_prop: bool
) -> vp.Document:
    """Swap the content between two layers

    This command swaps the content of layers FIRST and SECOND. Both FIRST and SECOND must be
    existing layer ids.

    The `--prob` option controls the probability with which each path are swapped. With a value
    lower than 1.0, some paths will remain in their original layer.

    If  the `--prob` option is not used, the layer properties are swapped between layers as
    well. This behaviour can be disabled with the `--no-prop` option.
    """

    first_lid = single_to_layer_id(first, document, must_exist=True)
    second_lid = single_to_layer_id(second, document, must_exist=True)

    if prob is None:
        document.swap_content(first_lid, second_lid)
        if not no_prop:
            document.layers[first_lid].metadata, document.layers[second_lid].metadata = (
                document.layers[second_lid].metadata,
                document.layers[first_lid].metadata,
            )
    else:
        new_first = vp.LineCollection()
        new_second = vp.LineCollection()

        for line in document.layers[first_lid]:
            (new_second if random.random() < prob else new_first).append(line)
        for line in document.layers[second_lid]:
            (new_first if random.random() < prob else new_second).append(line)

        document.replace(new_first, first_lid)
        document.replace(new_second, second_lid)

    return document


@cli.command(group="Layers")
@click.argument("layers", type=LayerType(accept_multiple=True, accept_new=False))
@global_processor
def lreverse(document: vp.Document, layers) -> vp.Document:
    """Reverse the path order within one or more layers.

    This command reverses the order in which paths are ordered within layer(s) LAYERS. LAYERS
    may be a single layer ID, multiple layer IDs (coma-separated without whitespace) or `all`
    (to refer to every exising layers).

    Examples:

        Delete layer one:

            $ vpype [...] ldelete 1 [...]

        Delete layers 1 and 2:

            $ vpype [...] ldelete 1,2 [...]

        Delete all layers:

            $ vpype [...] ldelete all [...]
    """

    lids = set(multiple_to_layer_ids(layers, document))

    for layer_id in lids:
        document.layers[layer_id].reverse()

    return document



"""
This code is adapted from https://github.com/fogleman/axi, which comes with the following
notice:

Copyright (C) 2017 Michael Fogleman

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import itertools
import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Callable

from .model import LineCollection

__all__ = ["FONT_NAMES", "text_line", "text_block"]


_FONT_DIR = Path(__file__).parent / "fonts"

FONT_NAMES = [p.stem for p in _FONT_DIR.glob("*.pickle")]


@dataclass
class _Glyph:
    """Glyph description class."""

    lt: float
    rt: float
    lines: LineCollection

    def append_with_offset(self, target: LineCollection, dx: float, dy: float = 0.0) -> None:
        """Append the glyph to a target LineCollection instance using the provided offset."""
        for line in self.lines:
            target.append(line + complex(dx, dy))


class _Font:
    """Font description class.

    Loads pickled version of the Ax i's Hershey font in:
    https://github.com/fogleman/axi/blob/master/axi/hershey_fonts.py

    The glyph data is converted to LineCollection.
    """

    _FONTS: dict[str, _Font] = {}

    def __init__(self, name: str):
        with open(_FONT_DIR / (name + ".pickle"), "rb") as fp:
            font_data = pickle.load(fp)

        # convert all glyphs to LineCollections
        self.glyphs: list[_Glyph] = [
            _Glyph(
                lt, rt, LineCollection([[complex(i, j) for i, j in line] for line in lines])
            )
            for lt, rt, lines in font_data
        ]

        # compute base height for this font
        all_glyphs = LineCollection()
        for glyph in self.glyphs:
            all_glyphs.extend(glyph.lines)
        self.max_height = all_glyphs.height()

    @classmethod
    def get(cls, font_name: str) -> _Font:
        """Get _Font instance for provided name, using caching."""
        try:
            if font_name not in cls._FONTS:
                cls._FONTS[font_name] = _Font(font_name)
            return cls._FONTS[font_name]
        except FileNotFoundError as exc:
            raise ValueError(f"font '{font_name}' could not be found", exc)


def _text_line(
    txt: str, font: _Font, spacing: float = 0.0, extra: float = 0.0
) -> LineCollection:
    """Generate a single line of text."""
    result = LineCollection()
    x = 0.0
    for ch in txt:
        index = ord(ch) - 32
        if index < 0 or index >= 96:
            x += spacing
            continue
        glyph = font.glyphs[index]
        glyph.append_with_offset(result, x - glyph.lt)
        x += glyph.rt - glyph.lt + spacing
        if index == 0:
            x += extra

    return result


def text_line(
    txt: str,
    font_name="futural",
    size: float = 18.0,
    align: str = "left",
    spacing: float = 0.0,
) -> LineCollection:
    """Create a line of text.

    The text block starts at (0, 0) and extends either right, left, or both, depending on the
    ``align`` argument.

    Args:
        txt: the text to layout
        font_name: the name of the font to use (see :const:`FONT_NAMES`)
        size: the font size
        align: text horizontal alignment (``"left"``, ``"right"``, or ``"center"``, default:
            left alignment)
        spacing: additional spacing added after each character
    """
    font = _Font.get(font_name)
    lc = _text_line(txt, font, spacing)
    lc.scale(size / font.max_height)

    if align == "right":
        lc.translate(-lc.width(), 0)
    elif align == "center":
        lc.translate(-lc.width() / 2, 0)
    elif align != "left":
        raise ValueError(f"unknown value for align ('{align}')")

    return lc


def _word_wrap(paragraph: str, width: float, measure_func: Callable[[str], float]):
    """Break text in multiple line."""
    result = []
    for line in paragraph.split("\n"):
        # handle empty lines
        if line == "":
            result.append("\n")
            continue

        fields = itertools.groupby(line, lambda c: c.isspace())
        fields = ["".join(g) for _, g in fields]
        if len(fields) % 2 == 1:
            fields.append("")
        x = ""
        for a, b in zip(fields[::2], fields[1::2]):
            w = measure_func(x + a)
            if w > width:
                if x == "":
                    result.append(a)
                    continue
                else:
                    result.append(x)
                    x = ""
            x += a + b
        if x != "":
            result.append(x + "\n")
    return result


def _justify_text(txt: str, font: _Font, width: float) -> LineCollection:
    """Draw text with justification."""
    txt = txt.strip()
    d = _text_line(txt, font)
    bounds = d.bounds()
    w = bounds[2] if bounds else 0.0
    spaces = txt.count(" ")
    if spaces == 0 or w >= width:
        return d
    e = (width - w) / spaces
    return _text_line(txt, font, extra=e)


def text_block(
    paragraph: str,
    width: float,
    font_name="futural",
    size: float = 18.0,
    align: str = "left",
    line_spacing: float = 1,
    justify=False,
) -> LineCollection:
    """Create a wrapped block of text using the provided width.

    The text block starts at (0, 0) and extends right and down. The parameters affects how the
    text is rendered.

    Args:
        paragraph: the text to layout
        font_name: the name of the font to use (see :const:`FONT_NAMES`)
        width: the width of the block
        size: the font size
        align: text horizontal alignment (``"left"``, ``"right"``, or ``"center"``, default:
            left alignment)
        line_spacing: line spacing (default: 1.0)
        justify: should the text be justified (default: False)
    """

    font = _Font.get(font_name)
    scale_factor = size / font.max_height
    width /= scale_factor

    def measure(txt):
        # noinspection PyShadowingNames
        bounds = _text_line(txt, font).bounds()
        return bounds[2] if bounds else 0.0

    lines = _word_wrap(paragraph, width, measure)

    lc_arr = [
        _justify_text(line, font, width)
        if justify and not line.endswith("\n")
        else _text_line(line, font)
        for line in lines
    ]

    spacing = line_spacing * font.max_height
    result = LineCollection()
    y = 0.0
    for lc in lc_arr:
        bounds = lc.bounds()
        w = bounds[2] if bounds else 0.0
        if align == "left":
            x = 0.0
        elif align == "right":
            x = width - w
        elif align == "center":
            x = width / 2 - w / 2
        else:
            raise ValueError(f"unknown value for align ('{align}')")

        for line in lc:
            result.append(line + complex(x, y))
        y += spacing

    result.scale(scale_factor)
    return result
